"""The package benchmark is designed to benchmark.

The package contains:

- The function benchmark in the module benchmark.
  This function can be used to run multiple algorithms on multiple problems for
  multiple runs

- The module statistics. This module contains several functions to get
  statistical data from the results of the function benchmark.

"""
