
def pass_func(*args):
    """A function that does nothing.

    Parameters
    ----------
    args
        Variable length argument list, it does not matter what or how many
        parameters you use. Does not accept keyword arguments.

    """

    pass
